import { Server as HttpServer } from "http";
import { ServiceBase } from "./ServiceBase";
import { Server } from "socket.io";
import { SessionManager } from "../managers/SessionManager";

export class SocketIoService  extends ServiceBase{
  io: Server
  clients: Map<string, any> = new Map();

  constructor(private httpServer: HttpServer) {
    super()
    this.io = this.socketIo(httpServer);   
    this.initListeners();
  }

  public getServer(): Server {
    return this.io;
  }

  private initListeners() {
    const sessionController = new SessionManager();
    this.io.on('connection', (socket) => {
      this.logger.debug(`connect ${socket.id}`);
      if (socket.recovered) {
        // recovery was successful: socket.id, socket.rooms and socket.data were restored
        this.logger.debug("recovered!");
        this.logger.debug("socket.rooms:", socket.rooms);
        this.logger.debug("socket.data:", socket.data);
      } else {
        this.logger.debug("new connection");
        this.clients.set(socket.id, { alive: true });
        socket.join('room-general')
        socket.data.foo = "bar";
      }
      
      socket.on('pong', () => {
        if (this.clients.has(socket.id)) {
          this.clients.set(socket.id, { alive: true });
        }
      })
    
      socket.on('disconnect', () => {
        this.logger.debug('user disconnected');
        this.clients.delete(socket.id);
      });
    
      socket.on('createSession', (data, callback) => {
        const response = sessionController.createSession(data, socket.id);
        callback(response);
      });

      socket.on('startSession', (data, callback) => {
        const response = sessionController.startSession(data);    
        callback(response);
      });
    
       socket.on('joinSession', (data, callback) => {
        const response = sessionController.joinSession(data, socket.id);
        callback(response);
      });

      socket.on('playerReady', (data, callback) => {
        const response = sessionController.setPlayerReady(data);
        callback(response);
      });
      
      this.pingClients()
    });
  }

  private pingClients() {
    setInterval(() => {
      for (let [id, client] of this.clients.entries()) {
        if (!client.alive) {
          this.logger.debug(`Client ${id} did not respond. Disconnecting.`);
          this.io.to(id).disconnectSockets(true); // Disconnect client
          this.clients.delete(id);
        } else {
          client.alive = false; // Reset alive status for the next ping
          this.io.to(id).emit('ping'); // Send ping message
        }
      }
    }, 30000); 
  }
  
  private socketIo(httpServer: HttpServer): Server {
    return new Server(httpServer, {
      cors: {
        origin: '*',
      },
      connectionStateRecovery: {
        maxDisconnectionDuration: 15 * 60 * 1000,
        skipMiddlewares: true,
      },
      connectTimeout: 15 * 60 * 1000,
    })
  }
}   